"""
Specifies and implements the MDF the function ontology; a collection of builtin functions that can be used in
MDF :class:`~modeci.mdf.Function` and :class:`~modeci.mdf.Parameter` objects. Code for registering standard functions
are with the ontology is implemented in :mod:`~modeci_mdf.functions.standard`. If you want to add functions to the
ontology that can be used during execution see how this has been done for ONNX (:mod:`~modeci_mdf.functions.onnx`) and
ACT-R (:mod:`~modeci_mdf.functions.actr`)
"""
